/******************************************************************************
 * The copyright in this software is being made available under the BSD        *
 * License, included below. This software may be subject to other third party  *
 * and contributor rights, including patent rights, and no such rights are     *
 * granted under this license.                                                 *
 *    <OWNER> = ISO                                                            *
 *    <ORGANIZATION> = ISO                                                     *
 *    <YEAR> = 2015                                                            *
 *    Copyright (c) 2012-2015, ISO                                             *
 *                                                                             *
 * All rights reserved.                                                        *
 * Redistribution and use in source and binary forms, with or without          *
 * modification, are permitted provided that the following conditions are met: *
 *       * Redistributions of source code must retain the above copyright      *
 *         notice,this list of conditions and the following disclaimer.        *
 *       * Redistributions in binary form must reproduce the above             *
 *         copyright notice, this list of conditions and the following         *
 *         disclaimer in the documentation and/or other materials provided     *
 *         with the distribution.                                              *
 *       * Neither the name of the <ORGANIZATION> nor the names of its         *
 *         contributors may be used to endorse or promote products derived     *
 *         from this software without specific prior written permission.       *
 *                                                                             *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" *
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE   *
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE  *
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE   *
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR         *
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF        *
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS    *
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN     *
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)     *
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE  *
 * POSSIBILITY OF SUCH DAMAGE.                                                 *
 ******************************************************************************/

#ifndef _JH_HELPERS_H
#define _JH_HELPERS_H

#include <stdio.h>
#include <stdarg.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>
#include "jh_host.h"
#include "jh_datatypes.h"
#include "jh_malloc.h"

#ifdef __cplusplus
extern "C" {
#endif
    
#define JH_HIST_SIZE      1025
#define JH_LN_MAX_LUM     ((float)(10 * log(10)))
#define JH_LN_JH_MIN_LUM  (-JH_LN_MAX_LUM)
#define JH_MAX_DR         (JH_LN_MAX_LUM - JH_LN_JH_MIN_LUM)
#define JH_MIN_LUM        (expf(JH_LN_JH_MIN_LUM))
#define JH_MAX_LUM        (expf(JH_LN_MAX_LUM))

/* JH_BITS = 13 is for showing images on large and bright TVs
 * JH_BITS = 11 should be enough for mobile devices
 */
#define	JH_BITS (12)

/* Max tonemapped luminance is jh_hdr_lum(RGB) where RGB is the max tonemapped RGB mapped to HDR color space, and is expected to
 * be (1.0, 1.0, 1.0).  Similarly, min tonemapped luminance is expected to result from converting RGB = (0,0,0) through the bga
 * curve and csb matrix, and computing luminance.  By default this will be equal to the min value in bga, (1.0f / (1 << JH_BITS)).
 */
#define JH_MAX_TONEMAP_LUM 1.0f
#define JH_MIN_TONEMAP_LUM (1.0f / (1 << JH_BITS))


    inline static unsigned char jh_byteclip(float val) {
        if (val < 1.f)
            return 0x00;
        if(val >= 255.f)    /* Parsing in int is slower at least for A9 arch */
            return 0xff;   /* for ex. exp > JH_BITS || (exp == JH_BITS & matissa > 1.) */
        return  (unsigned char)val;
    }

    inline static unsigned short jh_round(float val) {
        int *a = (int*)&val, e = 0;
        if (val < (float)(0.5 / (JH_BITS - 1))) {
            return 0;
        }
        if (val >= (1 << JH_BITS) - 1) { /* Parsing in int is slower at least for A9 arch */
            return (1 << JH_BITS) - 1;   /* for ex. exp > JH_BITS || (exp == JH_BITS & matissa > 1.) */
        }
        e = 22 - ((*a >> 23) - 127);
        *a = (((*a & 0x7fffff) | 0x800000) >> (e));
        return (unsigned short)((++(*a)) >> 1);
    }
    
    /*! \brief inplace cumulative sum of a floating point array
        \param[in] in - pointer to an float array
        \param[in] els - number of elements in the array
        \note no null pointer check is done
     */
    inline static void jh_cumulative_sum(float* in, const size_t els) {
        float* end = in + els;
        ++in; /*first element should not be summed*/
        while(in < end) {
            *in += *(in - 1);
            ++in;
        }
    }
    
    inline static int jh_invert_3x3_matrix(const float mat[9], float inv_mat[9])
    {
        float t1, t2, t3, det, invdet;
        t1 = mat[4]*mat[8] - mat[5]*mat[7];
        t2 = mat[5]*mat[6] - mat[3]*mat[8];
        t3 = mat[3]*mat[7] - mat[4]*mat[6];
        det = mat[0] * t1 + mat[1] * t2 + mat[2] * t3;
        if(det < 0.000001f) {
            return 1;
        }
        invdet = (float)1.0/(det);
        inv_mat[0] = t1 * invdet;
        inv_mat[1] = (mat[2]*mat[7] - mat[1]*mat[8]) * invdet;
        inv_mat[2] = (mat[1]*mat[5] - mat[2]*mat[4]) * invdet;
        inv_mat[3] = t2 * invdet;
        inv_mat[4] = (mat[0]*mat[8] - mat[6]*mat[2]) * invdet;
        inv_mat[5] = (mat[2]*mat[3] - mat[0]*mat[5]) * invdet;
        inv_mat[6] = t3 * invdet;
        inv_mat[7] = (mat[1]*mat[6] - mat[0]*mat[7]) * invdet;
        inv_mat[8] = (mat[0]*mat[4] - mat[1]*mat[3]) * invdet;
        return 0;
    }

    inline static void* jh_get_row(const jh_image_t* im, int row) {
        return (void*)((char*)im->data + row * im->stride);
    }
    
    inline static void jh_delete_image(jh_image_t* im) {
        if (NULL != im) {
            jh_free(im->decoding_lut);
            jh_free(im->loglum);
            jh_free(im->data);
            jh_free(im);
        }
        return;
    }

    inline static unsigned int jh_channel_count(const jh_pixel_type_t pixel_type) {
        unsigned int channel_count = 0U;
        switch (pixel_type) {
            case JH_RGBF:
            case JH_RGB8:
            case JH_YUV444I8:
                channel_count = 3U;
                break;
            case JH_GRAYF:
            case JH_GRAY8:
                channel_count = 1U;
                break;
            default:
                break;
        };
        return(channel_count);
    }
    
    inline static size_t jh_pixel_size(const jh_pixel_type_t pixel_type)
    {
        size_t pixel_size = 0U;
        switch (pixel_type) {
            case JH_GRAYF:
            case JH_RGBF:
                pixel_size = sizeof(float);
                break;
            case JH_YUV444I8:
            case JH_RGB8:
            case JH_GRAY8:
                pixel_size = sizeof(unsigned char);
                break;
            default:
                break;
        };
        return(pixel_size);
    }
    
    inline static int jh_decoding_lut_size(const jh_pixel_type_t pixel_type)
    {
        int lut_size = 0;
        switch (pixel_type) {
            case JH_RGB8:
            case JH_GRAY8:
                lut_size = 256;
                break;
            default:
                break;
        }
        return(lut_size);
    }
    
    /* Gamma decoding */
    inline static float jh_gamma_decode(float x) { /* Assume sRGB gamma */
        if (x <= 0.03928f) {
            return x / 12.92f;
        }
        return powf((x + 0.055f) / 1.055f, 2.4f);
        /* return powf(x, 2.2f); -- for non-sRGB */
    }
    
    /* Gamma encoding */
    inline static float jh_gamma_encode(float x) { /* Assume sRGB gamma */
        if (x <= 0.00304f) {
            return 12.92f * x;
        }
        return 1.055f * powf(x, 1.f / 2.4f) - 0.055f;
        /* return powf(x, 1.f / 2.2f); -- for non-sRGB  */
    }
    
    inline static void jh_fill_decoding_lut(jh_image_t *img)
    {
        if (NULL != img->decoding_lut && img->lut_size >= 256) {
            int i;
            img->lut_size = 256;
            for (i = 0; i < 256; i++ ) {
                img->decoding_lut[i] = jh_gamma_decode(i/255.f) + 1.f/(1<<JH_BITS);
            }
        }
        return;
    }
    
    inline static jh_image_t* jh_create_image(const int columns, const int rows,
                                              const jh_pixel_type_t pixel_type)
    {
        jh_image_t* im = (jh_image_t*)jh_zalloc(sizeof(jh_image_t));
        int er = 0;
        size_t el_size = jh_pixel_size(pixel_type);
        unsigned int channels = jh_channel_count(pixel_type);
        if ((0 == rows) || (0 == columns) || (0 == el_size) || (0 == channels)  /* don't create empty image */
           || (NULL == im)) {
            er = 1;
            goto cleanup;
        }
        im->lut_size = jh_decoding_lut_size(pixel_type);
        if (im->lut_size) { /* malloc(0) is not defined thus if */
            im->decoding_lut = (float *) jh_malloc(im->lut_size*sizeof(float));
            if (NULL == im->decoding_lut) {
                er = 1;
                goto cleanup;
            }
            jh_fill_decoding_lut(im);
        }
        im->data = jh_malloc(rows * columns * channels * el_size);
        if (NULL != im->data) {
            im->columns = columns;
            im->rows = rows;
            im->pixel_type = pixel_type;
            im->stride = columns * channels * (unsigned int) el_size;
            im->loglum = NULL;
        } else {
            er = 1;
            goto cleanup;
        }
    cleanup:
        if (er) {
            jh_delete_image(im);
            im = NULL;
        }
        return im;
    }
    
    inline static int jh_test_and_reset(volatile int* cancel) {
        int c = *cancel;
        *cancel = 0;
        return c;
    }

    inline static void jh_set_error(jh_error_t *err, const jh_result_t res, const char* desc, ...) {
        va_list vars;
        va_start(vars, desc);
        err->code = res;
        vsnprintf(err->description, JH_MSG_SIZE, desc, vars);
        va_end(vars);
    }
    
    inline static int jh_is_invalid_error(jh_error_t *error) {
        if (error == NULL) {
            return 1;
        }
        
        return 0;
    }
    
    inline static int jh_is_invalid_image(const jh_image_t *img, jh_error_t *error) {
        if (img == NULL) {
            jh_set_error(error, JH_INVALID_ARGUMENT, "NULL image parameter");
            return 1;
        }
        switch (img->pixel_type) {
            case JH_RGB8:
            case JH_RGBF:
            case JH_YUV444I8:
            case JH_GRAYF:
            case JH_GRAY8:
                break;
            default:
                jh_set_error(error, JH_INVALID_ARGUMENT,
                             "Illegal value for pixel_type (%d)", img->pixel_type);
                return 1;
        }
        if (img->data == NULL) {
            jh_set_error(error, JH_INVALID_ARGUMENT, "NULL image data");
            return 1;
        }
        return 0;
    }
    
    inline static int jh_is_null(const void *p, char *p_name, jh_error_t *error) {
        if (p == NULL) {
            jh_set_error(error, JH_INVALID_ARGUMENT, "NULL parameter %s", p_name);
            return 1;
        }
        return 0;
    }
    
    inline static jh_image_t* jh_copy_image(const jh_image_t *img)
    {
        int err = 0;
        jh_image_t *cpy = jh_create_image(img->columns, img->rows, img->pixel_type);
        if(!cpy) {
            goto cleanup;
        }
        memcpy(cpy->data, img->data, img->columns * img->rows * jh_pixel_size(img->pixel_type) * jh_channel_count(img->pixel_type));
        if(cpy->lut_size) {
            /* decoding_lut is allocated in jh_create_image */
            memcpy(cpy->decoding_lut, img->decoding_lut, cpy->lut_size * sizeof(float));
        }
        /* copy loglum if present */
        if(img->loglum) {
            size_t loglum_size = ((img->columns * img->rows) + 2) * sizeof(float);  /* +2 for loglum min & max at end of array */
            cpy->loglum = (float *) jh_malloc(loglum_size);
            if(!cpy->loglum) {
                err = 1;
                goto cleanup;
            }
            memcpy(cpy->loglum, img->loglum, loglum_size);
        }
    cleanup:
        if (err) {
            jh_delete_image(cpy);
            cpy = NULL;
        }
        return cpy;
    }
    
    
#ifdef __cplusplus
}
#endif

#endif /*_JH_HELPERS_H*/
