/******************************************************************************
 * The copyright in this software is being made available under the BSD        *
 * License, included below. This software may be subject to other third party  *
 * and contributor rights, including patent rights, and no such rights are     *
 * granted under this license.                                                 *
 *    <OWNER> = ISO                                                            *
 *    <ORGANIZATION> = ISO                                                     *
 *    <YEAR> = 2015                                                            *
 *    Copyright (c) 2012-2015, ISO                                             *
 *                                                                             *
 * All rights reserved.                                                        *
 * Redistribution and use in source and binary forms, with or without          *
 * modification, are permitted provided that the following conditions are met: *
 *       * Redistributions of source code must retain the above copyright      *
 *         notice,this list of conditions and the following disclaimer.        *
 *       * Redistributions in binary form must reproduce the above             *
 *         copyright notice, this list of conditions and the following         *
 *         disclaimer in the documentation and/or other materials provided     *
 *         with the distribution.                                              *
 *       * Neither the name of the <ORGANIZATION> nor the names of its         *
 *         contributors may be used to endorse or promote products derived     *
 *         from this software without specific prior written permission.       *
 *                                                                             *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" *
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE   *
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE  *
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE   *
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR         *
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF        *
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS    *
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN     *
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)     *
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE  *
 * POSSIBILITY OF SUCH DAMAGE.                                                 *
 ******************************************************************************/

#ifndef JH_LOG_H
#define JH_LOG_H

/*  Look up for ICSILog algorithm description on the Internet.
    DOLBY_LOGN is a size for a mantissa mask, which will be used
    to index lookup table. It has to be in the range [0..23].
    The bigger it is the more precise value of JH_LOG we will get.
    The idea is if it's small enough the entire lookup table
    can end up in cache, and we will get a significant speed up.
*/
#include "jh_log_data.h"


/*  Computes an approximation of log(val) quickly.
    val is a IEEE 754 float value, must be >0.
    lookup_table and n must be the same values as
    provided to fill_icsi_table.
    returns: log(val). No input checking performed.
*/
inline static float jh_log (float fval)
{
  int *const exp_ptr = ((int*)&fval);
  int x = *exp_ptr; /*x is treated as integer*/
  const int log_2 = ((x >> 23) & 255) - 127; /*exponent*/
  x &= 0x7FFFFF; /*mantissa*/
  x = x >> (23-DOLBY_LOGN); /*quantize mantissa*/
  fval = dolby_log_lookup_table[x]; /*lookup precomputed value*/
  return (float) ((fval + log_2)* 0.69314718); /*natural logarithm*/
}

/* Computes an approximation of log2(val).
   val is a IEEE 754 float value, must be >0.
   lookup_table and n must be the same values as
   provided to fill_icsi_table.
   returns: log(val). No input checking performed.
*/
inline static float jh_log2 (float fval)
{
  int *const exp_ptr = ((int*)&fval);
  int x = *exp_ptr; /*x is treated as integer*/
  const int log_2 = ((x >> 23) & 255) - 127; /*exponent*/
  x &= 0x7FFFFF; /*mantissa*/
  x = x >> (23-DOLBY_LOGN); /*quantize mantissa*/
  fval = dolby_log_lookup_table[x]; /*lookup precomputed value*/
  return (float) (fval + log_2); /*natural logarithm*/
}

#endif /*JH_LOG_H*/
