/******************************************************************************
 * The copyright in this software is being made available under the BSD        *
 * License, included below. This software may be subject to other third party  *
 * and contributor rights, including patent rights, and no such rights are     *
 * granted under this license.                                                 *
 *    <OWNER> = ISO                                                            *
 *    <ORGANIZATION> = ISO                                                     *
 *    <YEAR> = 2015                                                            *
 *    Copyright (c) 2012-2015, ISO                                             *
 *                                                                             *
 * All rights reserved.                                                        *
 * Redistribution and use in source and binary forms, with or without          *
 * modification, are permitted provided that the following conditions are met: *
 *       * Redistributions of source code must retain the above copyright      *
 *         notice,this list of conditions and the following disclaimer.        *
 *       * Redistributions in binary form must reproduce the above             *
 *         copyright notice, this list of conditions and the following         *
 *         disclaimer in the documentation and/or other materials provided     *
 *         with the distribution.                                              *
 *       * Neither the name of the <ORGANIZATION> nor the names of its         *
 *         contributors may be used to endorse or promote products derived     *
 *         from this software without specific prior written permission.       *
 *                                                                             *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" *
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE   *
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE  *
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE   *
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR         *
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF        *
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS    *
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN     *
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)     *
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE  *
 * POSSIBILITY OF SUCH DAMAGE.                                                 *
 ******************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "jh_datatypes.h"
#include "jh_helpers.h"
#include "jh_checksum.h"

/* ------------ Checksum structure and calculation ----------------------------- */

typedef struct jh_fletcher16_checksum_ts
{
    unsigned char C0;
    unsigned char C1;
} jh_fletcher16_checksum_t;

static void jh_checksum_init(jh_fletcher16_checksum_t* cs)
{
    cs->C0 = 0;
    cs->C1 = 0;
    return;
}

static void jh_checksum_add_data(jh_fletcher16_checksum_t* cs, const unsigned char *data, const unsigned long numBytes)
{
    unsigned long i;
    unsigned short sum = 0;
    for(i=0; i<numBytes; i++) {
        sum  = cs->C0 + *data++;
        cs->C0 = (unsigned char) (sum + ((sum + 1) >> 8));
        sum  = (cs->C0 + cs->C1);
        cs->C1 = (unsigned char) (sum + ((sum + 1) >> 8));
    }
    return;
}

unsigned short jh_checksum_finalize(jh_fletcher16_checksum_t* cs)
{
    return (cs->C1 << 8) | cs->C0;
}

/* ------------ Application of checksum to JPEG data --------------------------- */

/* From Appendix F of the spec:
 
 F.1	Data Covered by the Checksum (normative)
 
 The checksum algorithm, specified in subclause F.2 is to be computed over the
 following data:
 The checksum starts with and includes the first SOF marker of the legacy
 codestream, starting with the marker code itself, and proceeds until the last
 byte of the entropy coded segments of the legacy codestream. However, the
 following markers in the Tables/misc section upfront the frame header are not
 to be included in the checksum computation:
 •	Any APPn marker, the marker indiciator, their length and their contents is
 to be skipped for the purpose of the checksum computation.
 •	Any COM marker, the marker indicator, its length and its contents is to be
 skipped.
 •	The EOI marker shall not be included in the computation of the checksum.
 
 NOTE – By this definition, RSTn (restart markers) shall be included in the
 computation of the checksum.
 
 */


/* JPEG marker codes */
typedef enum {
    
    /* 0xFF00, 0xFF01, 0xFFFE, 0xFFC0-0xFFDF(?) ITU T.81/IEC 10918-1 */
    /* 0xFFF0 - 0xFFF6 ITU T.84/IEC 10918-3 JPEG extensions */
    /* 0xFFF7 - 0xFFF8 ITU T.87/IEC 14495-1 JPEG LS */
    /* 0xFF4F - 0xFF6f, 0xFF90 - 0xFF93 JPEG 2000 IEC 15444-1 */
    /* 0xFF30 - 0xFF3F reserved JP2 (markers only--no marker segments */
    
    /* JPEG 1994 - defined in ITU T.81 | ISO IEC 10918-1 */
    SOF0  = 0xc0,       /* start of frame - baseline jpeg */
    SOF1  = 0xc1,       /* extended sequential, huffman */
    SOF2  = 0xc2,       /* progressive, huffman */
    SOF3  = 0xc3,       /* lossless, huffman */
    
    SOF5  = 0xc5,       /* differential sequential, huffman */
    SOF6  = 0xc6,       /* differential progressive, huffman */
    SOF7  = 0xc7,       /* differential lossless, huffman */
    JPG   = 0xc8,       /* reserved for JPEG extension */
    SOF9  = 0xc9,       /* extended sequential, arithmetic */
    SOF10 = 0xca,       /* progressive, arithmetic */
    SOF11 = 0xcb,       /* lossless, arithmetic */
    
    SOF13 = 0xcd,       /* differential sequential, arithmetic */
    SOF14 = 0xce,       /* differential progressive, arithmetic */
    SOF15 = 0xcf,       /* differential lossless, arithmetic */
    
    DHT   = 0xc4,       /* define huffman tables */
    
    DAC   = 0xcc,       /* define arithmetic-coding conditioning */
    
    /* restart markers */
    RST0  = 0xd0,
    RST1  = 0xd1,
    RST2  = 0xd2,
    RST3  = 0xd3,
    RST4  = 0xd4,
    RST5  = 0xd5,
    RST6  = 0xd6,
    RST7  = 0xd7,
    /* delimeters */
    SOI   = 0xd8,       /* start of image */
    EOI   = 0xd9,       /* end of image */
    SOS   = 0xda,       /* start of scan */
    DQT   = 0xdb,       /* define quantization tables */
    DNL   = 0xdc,       /* define number of lines */
    DRI   = 0xdd,       /* define restart interval */
    DHP   = 0xde,       /* define hierarchical progression */
    EXP   = 0xdf,       /* expand reference components */
    
    /* JPEG 1997 extensions ITU T.84 | ISO IEC 10918-3 */
    /* application data sections */
    APP0  = 0xe0,
    APP1  = 0xe1,
    APP2  = 0xe2,
    APP3  = 0xe3,
    APP4  = 0xe4,
    APP5  = 0xe5,
    APP6  = 0xe6,
    APP7  = 0xe7,
    APP8  = 0xe8,
    APP9  = 0xe9,
    APP10 = 0xea,
    APP11 = 0xeb,
    APP12 = 0xec,
    APP13 = 0xed,
    APP14 = 0xee,
    APP15 = 0xef,
    /* extention data sections */
    JPG0  = 0xf0,
    JPG1  = 0xf1,
    JPG2  = 0xf2,
    JPG3  = 0xf3,
    JPG4  = 0xf4,
    JPG5  = 0xf5,
    JPG6  = 0xf6,
    SOF48 = 0xf7,       /* JPEG-LS */
    LSE   = 0xf8,       /* JPEG-LS extension parameters */
    JPG9  = 0xf9,
    JPG10 = 0xfa,
    JPG11 = 0xfb,
    JPG12 = 0xfc,
    JPG13 = 0xfd,
    JCOM  = 0xfe,       /* comment */
    
    TEM   = 0x01,       /* temporary private use for arithmetic coding */
    
    /* 0x02 -> 0xbf reserved in JPEG 94/97*/
    
    /* JPEG 2000 - defined in IEC 15444-1 "JPEG 2000 Core (part 1)" */
    /* delimiters */
    SOC   = 0x4f,	/* start of codestream */
    SOT   = 0x90,	/* start of tile */
    SOD   = 0x93,	/* start of data */
    EOC   = 0xd9,	/* end of codestream */
    /* fixed information segment */
    SIZ   = 0x51,	/* image and tile size */
    /* functional segments */
    COD   = 0x52,	/* coding style default */
    COC   = 0x53,	/* coding style component */
    RGN   = 0x5e,	/* region of interest */
    QCD   = 0x5c,   /* quantization default */
    QCC   = 0x5d,	/* quantization component */
    POC   = 0x5f,	/* progression order change */
    /* pointer segments */
    TLM   = 0x55,	/* tile-part lengths */
    PLM   = 0x57,	/* packet length (main header) */
    PLT   = 0x58,	/* packet length (tile-part header) */
    PPM   = 0x60,	/* packed packet headers (main header) */
    PPT   = 0x61,	/* packet packet headers (tile-part header) */
    /* bitstream internal markers and segments */
    SOP   = 0x91,	/* start of packet */
    EPH   = 0x92,	/* end of packet header */
    /* informational segments */
    CRG   = 0x63,	/* component registration */
    COM   = 0x64,	/* comment */
    
} marker;

static int jh_is_zero_length_segment(const int markerID)
{
    switch(markerID) {
        default:
            break;
        case SOI:
        case EOI:
        case RST0:
        case RST1:
        case RST2:
        case RST3:
        case RST4:
        case RST5:
        case RST6:
        case RST7:
            return(1);
            break;
    }
    return (0);
}

static int jh_is_start_of_frame(unsigned char marker_id)
{
    switch(marker_id) {
        case SOF0:
        case SOF1:
        case SOF2:
        case SOF3:
        case SOF5:
        case SOF6:
        case SOF7:
        case SOF9:
        case SOF10:
        case SOF11:
        case SOF13:
        case SOF14:
        case SOF15:
            return(1);
            break;
        default:
            break;
    }
    return(0);
}

static int jh_is_app_marker(unsigned char marker_id)
{
    switch(marker_id) {
        case APP0:
        case APP1:
        case APP2:
        case APP3:
        case APP4:
        case APP5:
        case APP6:
        case APP7:
        case APP9:
        case APP10:
        case APP11:
        case APP13:
        case APP14:
        case APP15:
            return(1);
            break;
        default:
            break;
    }
    return(0);
}

/* No APPn or COM marker segments are included in checksum */
static int jh_include_segment_type_in_checksum(unsigned char marker_id)
{
    if(jh_is_app_marker(marker_id) || (COM == marker_id)) {
        return(0);
    }
    return(1);
}

/*  Dolby APP11 Segments:
 *  Common Identifier: 2 bytes; value: DD
 *  Payload type:      2 bytes; values: 01 = parameters
 *                                      02 = residual image
 */
enum eAPP11ID { eDolbyAPP11ParametersID, eDolbyAPP11ResidualID, eUnknownAPP11ID };

static enum eAPP11ID jh_get_app11_id(const unsigned char *data)
{
    if((data[0]=='D') && (data[1]=='D')) {
        if((data[2]==0) && (data[3]==1)) {
            return(eDolbyAPP11ParametersID);
        }
        else if((data[2]==0) && (data[3]==2)) {
            return(eDolbyAPP11ResidualID);
        }
    }
    return(eUnknownAPP11ID);
}

/* Read a marker segment from stream, populating segment with its data */
static int jh_read_segment(jh_marker_segment_t *segment, jh_stream_t *jstream)
{
    unsigned char b1, b2;
    jstream->read(jstream, &b1, sizeof(unsigned char));
    if(b1 != 0xFF) {
        /* stream not properly positioned, or we're in a scan */
        return 0;
    }
    
    /* read marker ID */
    jstream->read(jstream, &(segment->marker), sizeof(unsigned char));
    
    /* handle zero-length markers */
    switch(segment->marker) {
        default:
            break;
        case 0x00:
        case 0xFF:
        case SOI:
        case EOI:
        case RST0:
        case RST1:
        case RST2:
        case RST3:
        case RST4:
        case RST5:
        case RST6:
        case RST7:
            segment->length = 0;
            return(1);
    }
    
    /* read segment data */
    jstream->read(jstream, &b1, sizeof(unsigned char));
    jstream->read(jstream, &b2, sizeof(unsigned char));
    segment->length = b1 * 256 + b2;
    segment->length -= 2;  /* jh_marker_segment_t's length doesn't include the 2 byte length */
    jstream->read(jstream, segment->data, segment->length * sizeof(unsigned char));
    return 1;
}


/* Find the first occurrence of pattern in str, limiting the search to the first n characters.  This is missing from
 * some Linux distributions and so included here.
 */
static char *jh_strnstr(const char *str, const char *pattrn, size_t n)
{
    size_t len;
    char c, sc;
    if ((c = *pattrn++) != '\0') {
        len = strlen(pattrn);
        do {
            do {
                if ((n-- < 1) || ((sc = *str++) == '\0')) {
                    return (NULL);
                }
            } while (sc != c);
            if (len > n) {
                return (NULL);
            }
        } while (strncmp(str, pattrn, len) != 0);
        --str;
    }
    return ((char *)str);
}


/* If the segment contains a Dolby APP11 parameter segment, go inside and extract the checksum and offset to it.
 * Returns 1 if a valid checksum was located in the segment (implying it was an APP11 parameter segment and it
 * didn't contain just a placeholder checksum value) and 0 otherwise.
 */
int jh_get_checksum_value_and_offset(jh_marker_segment_t* segment, unsigned short *header_checksum, long long *offset_in_segment)
{
    /* parse the data, looking for "ckb=" */
    const char cksum[] = "ckb=";
    /* Need to skip over the beginning 8 bytes containing "DD", 0x00, major & minor version numbers, and number of data segments
     * entries in the header and get to the text string, otherwise jh_strnstr() will stop too early. */
    const unsigned int start_byte = 8;
    char * pCksumStr = NULL, *pCksum = NULL;
    if(eDolbyAPP11ParametersID != jh_get_app11_id(segment->data)) {
        return(0);
    }
    pCksumStr = (char *) jh_strnstr((const char *)&(segment->data[start_byte]), cksum, segment->length - start_byte);
    if(NULL == pCksumStr) {
        return(0);  /* search string not found */
    }
    
    /* found it; set the offset from the beginning of the segment (including 4 bytes for the ID and length) */
    *offset_in_segment = pCksumStr - (char *) (segment->data) + strlen(cksum) + 4;
    
    /* extract the bytes and compute it's value */
    pCksum = pCksumStr + strlen(cksum);  /* point to the 4 bytes that follow the ckb= label */
    if(*pCksum != '*') {  /* "****" is used as a checksum placeholder; do nothing more if that's what's in the header */
        /* not a placeholder; convert the 4 bytes into an unsigned int */
        sscanf(pCksum, "%4hX", header_checksum);
    }
    return(1);
}


/* add the contribution of a marker segment to the checksum */
static void jh_checksum_add_segment(jh_fletcher16_checksum_t* checksum, jh_marker_segment_t* segment)
{
    /* add the bytes for the marker ID and, if not a zero length marker, for the length and payload */
    unsigned char data[4];
    unsigned int adjusted_length = 0U;
    data[0] = 0xFF;
    data[1] = segment->marker;
    if(jh_is_zero_length_segment(segment->marker)) {
        jh_checksum_add_data(checksum, data, 2);
        return;
    }
    adjusted_length = segment->length + 2;  /* length in struct is 2 less than what's in the file */
    data[2] = (adjusted_length >> 8) & 0x000000FF;
    data[3] = adjusted_length & 0x000000FF;
    jh_checksum_add_data(checksum, data, 4);
    jh_checksum_add_data(checksum, segment->data, segment->length);
}



/* jh_compute_checksum() scans a stream and computes the checksum according to Appendix F.
 *
 * If extracted_checksum != NULL, then if a Dolby APP11 parameter segment is encountered during
 * the scan, it will be searched for a valid checksum entry ("ckb") which will be returned in
 * *extracted_checksum.  This is usually done when decoding.
 *
 * If checksum_offset != NULL, then if a Dolby APP11 parameter segment is encountered during
 * the scan, it will be searched for a valid checksum entry ("ckb"), and the offset into the
 * stream at which the 4 bytes of the checksum are located will be returned in *checksum_offset.
 * This is usually done when encoding.
 *
 * Returns the computed checksum value.
 */
unsigned short jh_compute_checksum(jh_stream_t* stream,
                                unsigned short* extracted_checksum,
                                long long* checksum_offset,
                                jh_error_t* error)
{
    int ret = 1;
    long long cksum_offset = 0;
    long long cur_offset = 0;
    unsigned short header_checksum = 0;
    unsigned char was_SOF = 0;
    unsigned char foundEOI = 0;
    unsigned int block_size = 8192;
    int lastByteWasFF = 0;
    int bytes_read = 0;
    jh_fletcher16_checksum_t checksum;  /* the computed checksum */
    
    /* make a marker segment big enough to hold any allowable payload size */
    jh_marker_segment_t segment;
    unsigned char data[65535];  /* also used when reading scan data */
    segment.data = data;
    
    /* initialize for no errors */
    jh_set_error(error, JH_OK, "No error");

    jh_checksum_init(&checksum);

    /* rewind the stream */
    ret = stream->seek(stream, 0, JH_SEEK_SET);
    if (!ret) {
        jh_set_error(error, JH_USER_STREAM_ERROR, "Failed to rewind stream for checksum calculation.");
        goto cleanup;
    }
    
    /* Process:
       jh_read_segment() returns 0 (stop) when the first byte in the stream isn't 0xFF, which means
       the stream isn't properly positioned OR it's the beginning of scan data.  So use this process
       to skip segments until SOFn is encountered, then start grabbing bytes for the checksum until
       we get to the end.
    
       OK, let's start parsing
     
       read from the beginning of the stream, skipping markers until we get to SOF.  Along the way,
       if we encounter the Dolby APP11 parameter segment, either extract the checksum or make a note
       of it's location (or both).
     */
    while (jh_read_segment(&segment, stream)) {
        /* If it's an APP11 segment, open it up, looking for Dolby APP11 parameters */
        if(segment.marker == APP11) {
            long long offset_in_segment = 0;
            if(jh_get_checksum_value_and_offset(&segment, &header_checksum, &offset_in_segment)) {
                /* it was the parameter segment; compute the overall offset */
                cksum_offset = cur_offset + offset_in_segment;
            }
        }
        cur_offset += (jh_is_zero_length_segment(segment.marker)) ? 2 : segment.length + 4;  /* +4 for marker ID & length */
        
        /* if EOI, break */
        if(EOI == segment.marker) {
            break;
        }
        if(jh_is_start_of_frame(segment.marker)) {
            /* initialize the checksum calculation */
            was_SOF = 1;
            jh_checksum_add_segment(&checksum, &segment);
            break;
        }
    }

    /* we've broken out of the prior while loop because we're at EOI or poorly positioned or just found SOFn */
    if(!was_SOF) {
        /* wasn't a clean exit from the prior loop */
        jh_set_error(error, JH_CHECKSUM_PARSE_ERROR, "Poorly positioned stream in checksum calculation.");
        goto cleanup;
    }
    
    /* parse more segments until start of scan (SOS) */
    while (jh_read_segment(&segment, stream)) {
        /* Check any APP11 markers, just like in the while above */
        if(segment.marker == APP11) {
            long long offset_in_segment = 0;
            if(jh_get_checksum_value_and_offset(&segment, &header_checksum, &offset_in_segment)) {
                /* it was the parameter segment; compute the overall offset */
                cksum_offset = cur_offset + offset_in_segment;
            }
        }
        cur_offset += (jh_is_zero_length_segment(segment.marker)) ? 2 : segment.length + 4;  /* +4 for marker ID & length */

        /* if EOI, break */
        if(EOI == segment.marker) {
            break;
        }
    
        if(jh_include_segment_type_in_checksum(segment.marker)) {
            jh_checksum_add_segment(&checksum, &segment);
        }
        
        if(SOS == segment.marker) {
            /* we've encountered SOS; break out of loop reading segments */
            break;
        }
    }
    
    /* now read block of bytes and add them to the checksum until the EOI marker is reached */
    do {
        bytes_read = stream->read(stream, data, block_size * sizeof(unsigned char));
        if(0 == bytes_read) {
            /* this may occur if EOI is missing from end of the jpeg stream */
            break;
        }
        
        /* Handle the case in which 0xff was the last byte of the previous block */
        /* and EOI is the first byte of the this block. */
        if(lastByteWasFF) {
            if(data[0]==EOI) {
                break;  /* we're done; 0xFF wasn't added on last iteration and EOI not added now */
            }
            else {
                /* add 0xFF to checksum to compensate for skipping 0xFF at the end of the last block */
                unsigned char ff = 0xFF;
                jh_checksum_add_data(&checksum, &ff, 1);
            }
        }
        
        lastByteWasFF = (data[bytes_read-1]!=0xFF) ? 0 : 1;  /* set flag for next iteration */
        if(lastByteWasFF) {
            /* encountered case where last byte is 0xFF; add all but last byte to checksum */
            --bytes_read;
        }
        else if((data[bytes_read-2]== 0xFF) && (data[bytes_read-1]== EOI)) {
            /* encountered case where EOI is last 2 bytes read; adjust byte count and set exit flag */
            foundEOI = 1;
            bytes_read -= 2;
        }

        /* add up to the EOI to the checksum */
        jh_checksum_add_data(&checksum, data, bytes_read);
    } while(0 == foundEOI);
    
    /* done */
cleanup:
    if((error->code == JH_OK) && (0 == header_checksum) && (0 == cksum_offset)) {
        /* This should be treated more as a warning than error, since older files won't have the checksum. */
        jh_set_error(error, JH_HEADER_NO_CHECKSUM, "Checksum not present in header.");
    }
    if(checksum_offset) {
        *checksum_offset = cksum_offset;
    }
    if(extracted_checksum) {
        *extracted_checksum = header_checksum;
    }
    return(jh_checksum_finalize(&checksum));
}


/* Before decoding, checks whether the checksum stored in the Dolby APP11 parameter segment
 * matches the value computed for the base image according to Appendix F.
 *
 * Returns JH_OK if checksum matches and nonzero otherwise.
 */
int jh_is_checksum_valid(jh_stream_t* stream, jh_error_t* error)
{
    int ret = JH_OK;
    unsigned short extracted_checksum = 0;
    unsigned short computed_checksum = 0;
    computed_checksum = jh_compute_checksum(stream, &extracted_checksum, NULL, error);
    ret = stream->seek(stream, 0, JH_SEEK_SET);
    
    if((error->code == JH_OK) && (computed_checksum != extracted_checksum)) {
        ret = JH_INVALID_CHECKSUM;
        jh_set_error(error, JH_INVALID_CHECKSUM, "Computed checksum does not match value in header.");
    }
    else if (!ret) {
        ret = JH_USER_STREAM_ERROR;
        jh_set_error(error, JH_USER_STREAM_ERROR, "Failed to rewind stream after checksum calculation.");
    }
    else {
        ret = error->code;
    }
    return(ret);
}


/* After encoding, adds a checksum for the base image, computed according to Appendix F, to the */
/* Dolby APP11 parameter segment. */
/* */
/* Returns 0 if checksum matches nonzero on error. */
/* */
int jh_add_checksum(jh_stream_t* stream, jh_error_t* error)
{
    int ret = 0;
    long long checksum_offset = 0;
    char checksum_str[5];
    unsigned short computed_checksum = jh_compute_checksum(stream, NULL, &checksum_offset, error);
    if(error->code != JH_OK) {
        goto cleanup;
    }
    /* get the checksum characters */
    sprintf(checksum_str, "%04X", computed_checksum);
    /* rewind the stream to the location of the checksum and write 4 checksum characters */
    ret = stream->seek(stream, checksum_offset, JH_SEEK_SET);
    if (!ret) {
        jh_set_error(error, JH_USER_STREAM_ERROR, "Failed to rewind stream for writing checksum.");
        goto cleanup;
    }
    if(4 != stream->write(stream, (unsigned char *) &(checksum_str[0]), 4)) {
        jh_set_error(error, JH_CHECKSUM_WRITE_ERROR, "Failed to write checksum to header.");
    }
cleanup:
    return(error->code);
}
