/*
 * Radiance HDR image file I/O routines, based on Radiance open source code.
 *
 * This product includes Radiance software (http://radsite.lbl.gov/)
 * developed by the Lawrence Berkeley National Laboratory (http://www.lbl.gov/).
 * See the complete Radiance Software License at the bottom of this file.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <math.h>

#define  RED		0
#define  GRN		1
#define  BLU		2
#define  EXP		3
#define  COLXS		128	/* excess used for exponent */

#define  copycolr(c1,c2)	(c1[0]=c2[0],c1[1]=c2[1], \
                             c1[2]=c2[2],c1[3]=c2[3])

typedef unsigned char COLR[4];		/* red, green, blue, exponent */
typedef float        COLOR[3];	    /* red, green, & blue floats */

/* definitions for resolution header */
#define  XDECR			1
#define  YDECR			2
#define  YMAJOR			4


/* HDR reader */

/* read in an old colr scanline */
static int oldreadcolrs(COLR  *scanline, int  len, FILE  *fp)
{
	int  i, rshift = 0;
	
	while (len > 0) {
		scanline[0][RED] = (unsigned char) getc(fp);
		scanline[0][GRN] = (unsigned char) getc(fp);
		scanline[0][BLU] = (unsigned char) getc(fp);
		scanline[0][EXP] = (unsigned char) getc(fp);
        if (feof(fp) || ferror(fp)) {
			return(-1);
        }
		if (scanline[0][RED] == 1 &&
            scanline[0][GRN] == 1 &&
            scanline[0][BLU] == 1) {
			for (i = scanline[0][EXP] << rshift; i > 0; i--) {
				copycolr(scanline[0], scanline[-1]);
				scanline++;
				len--;
			}
			rshift += 8;
		} else {
			scanline++;
			len--;
			rshift = 0;
		}
	}
	return(0);
}

/* read in an encoded colr scanline */
static int freadcolrs(COLR  *scanline, int  len, FILE  *fp)
{
    int  i, j, code;
    /* determine scanline type */
    const int MINELEN = 8;	/* minimum scanline length for encoding */
    if (len < MINELEN) {
		return(oldreadcolrs(scanline, len, fp));
    }
    if ((i = getc(fp)) == EOF) {
		return(-1);
    }
	if (i != 2) {
		ungetc(i, fp);
		return(oldreadcolrs(scanline, len, fp));
	}
	scanline[0][GRN] = (unsigned char) getc(fp);
	scanline[0][BLU] = (unsigned char) getc(fp);
    if ((i = getc(fp)) == EOF) {
		return(-1);
    }
	if (scanline[0][GRN] != 2 || scanline[0][BLU] & 128) {
		scanline[0][RED] = 2;
		scanline[0][EXP] = i;
		return(oldreadcolrs(scanline+1, len-1, fp));
	}
    if ((scanline[0][BLU]<<8 | i) != len) {
		return(-1);		/* length mismatch! */
    }
    /* read each component */
    for (i = 0; i < 4; i++) {
	    for (j = 0; j < len; ) {
            if ((code = getc(fp)) == EOF) {
                return(-1);
            }
            if (code > 128) {	/* run */
                scanline[j++][i] = (unsigned char) getc(fp);
                for (code &= 127; --code; j++) {
                    scanline[j][i] = scanline[j-1][i];
                }
            } else {		/* non-run */
                while (code--) {
                    scanline[j++][i] = (unsigned char) getc(fp);
                }
            }
	    }
    }
	return(feof(fp) ? -1 : 0);
}

/* convert short to float color */
static void colr_color(COLOR  col, COLR  clr)
{
	double  f;
	
	if (clr[EXP] == 0)
		col[RED] = col[GRN] = col[BLU] = 0.0;
	else {
		f = ldexp(1.0, (int)clr[EXP]-(COLXS+8));
		col[RED] = (clr[RED] + 0.5)*f;
		col[GRN] = (clr[GRN] + 0.5)*f;
		col[BLU] = (clr[BLU] + 0.5)*f;
	}
}

static const int FMTSTRL = 7;

/* Copy the format value in s to r; return format value */
static void formatval(char  *r, char  *s)
{
	s += FMTSTRL;
	while (isspace(*s)) s++;
	if (!*s) {
        *r = '\0';
        return;
    }
    while(*s) {
        *r++ = *s++;
    }
    while (isspace(r[-1])) {
        r--;
    }
	*r = '\0';
}

/* Read header from fp, calling f(s,p) on each line */
static int getheader(FILE  *fp,
                     void  (*f)(char *, char *),
                     char  *p)
{
    const unsigned int MAXLINE = 512;
	char  buf[MAXLINE];
    
	for ( ; ; ) {
		buf[MAXLINE-2] = '\n';
        
		if (fgets(buf, sizeof(buf), fp) == NULL)
			return(-1);
		if (buf[0] == '\n')
			return(0);
		if (buf[MAXLINE-2] != '\n') {
			ungetc(buf[MAXLINE-2], fp);	/* prevent false end */
			buf[MAXLINE-2] = '\0';
		}
        if (f != NULL) {
			(*f)(buf, p);
        }
	}
    return 0;
}

struct check {
	FILE	*fp;
	char	fs[64];
};

/* check a header line for format info. */
static void mycheck(char  *s, char *c)
{
     struct check  *cp = (struct check *) c;
    char  FMTSTR[] = "FORMAT=";
	if (!strncmp(s,FMTSTR,FMTSTRL))
		formatval(cp->fs, s);
	else if (cp->fp != NULL)	/* don't copy format info. */
		fputs(s, cp->fp);
    return;
}

/*
 * Copymatch(pat,str) checks pat for wildcards, and
 * copies str into pat if there is a match (returning true).
 */
#define copymatch(pat, s)	(!strcmp(pat, s))

/*
 * Checkheader(fin,fmt,fout) returns a value of 1 if the input format
 * matches the specification in fmt, 0 if no input format was found,
 * and -1 if the input format does not match or there is an
 * error reading the header.  If fmt is empty, then -1 is returned
 * if any input format is found (or there is an error), and 0 otherwise.
 * If fmt contains any '*' or '?' characters, then checkheader
 * does wildcard expansion and copies a matching result into fmt.
 * Be sure that fmt is big enough to hold the match in such cases!
 * The input header (minus any format lines) is copied to fout
 * if fout is not NULL.
 */
static int checkheader(FILE *fin, const char *fmt, FILE *fout)
{
	struct check cdat;
    
	cdat.fp = fout;
	cdat.fs[0] = '\0';
    
	if (getheader(fin, mycheck, (char *) &cdat) < 0)
		return(-1);
    
	if (cdat.fs[0] != '\0')
		return(copymatch(fmt, cdat.fs) ? 1 : -1);
	return(0);
}

static int fgetresolu(		/* get x and y resolution */
               int  *xrp, int *yrp,
               FILE  *fp
               )
{
	char  buf[64], *xndx, *yndx, *cp;
    int  ord;
    
	if (fgets(buf, sizeof(buf), fp) == NULL)
		return(-1);
    
	xndx = yndx = NULL;
	for (cp = buf+1; *cp; cp++)
		if (*cp == 'X')
			xndx = cp;
		else if (*cp == 'Y')
			yndx = cp;
	if (xndx == NULL || yndx == NULL)
		return(-1);
	ord = 0;
	if (xndx > yndx) ord |= YMAJOR;
	if (xndx[-1] == '-') ord |= XDECR;
	if (yndx[-1] == '-') ord |= YDECR;
	if ((*xrp = atoi(xndx+1)) <= 0)
		return(-1);
	if ((*yrp = atoi(yndx+1)) <= 0)
		return(-1);
	return(ord);
}

static int getrow(FILE* in, float *fbuf, int xsize)
{
    COLOR   color;
    COLR    *scanin = (COLR*) malloc(xsize * sizeof(COLR));
    int	    x, k=0;
    
    if (scanin == NULL) {
        fprintf(stderr, "not enough memory\n");
        return(1);
    }
    
    if (freadcolrs(scanin, xsize, in) < 0) {
        fprintf(stderr, "read error\n");
        free(scanin);
        return(1);
    }
    
    for (x = 0; x < xsize; x++) {
        colr_color(color, scanin[x]);
        fbuf[k++] = color[RED];
        fbuf[k++] = color[GRN];
        fbuf[k++] = color[BLU];
	}

    free(scanin);
    return(0);
}

int read_radiance_hdr(const char* infile, float** hdrdata, unsigned int* width, unsigned int* height)
{
    int ret = 1;
    
    /* picture format identifier */
    const char *COLRFMT = "32-bit_rle_rgbe";
    const unsigned int RGB = 3;
    float *rgb = NULL;

    *width = 0;
    *height = 0;
    *hdrdata = NULL;

    FILE *fp = fopen(infile, "rb");
    if(NULL == fp) {
        fprintf(stderr, "Failed to open image file %s\n", infile);
        return ret;
    }
    
    /* read the size of the Radiance picture */
    int xsize, ysize;
    if(checkheader(fp, COLRFMT, NULL) < 0 ||
        fgetresolu(&xsize, &ysize, fp) != (YMAJOR|YDECR) ) {
        fprintf(stderr, "%s: not a radiance picture\n", infile);
        goto cleanup;
	}
    
    rgb = (float *) malloc(RGB * xsize * ysize * sizeof(float));
    if (rgb == NULL) {
        fprintf(stderr, "not enough memory\n");
        goto cleanup;
	}

    for(int y=0; y<ysize; y++) {
        if(1 == getrow(fp, &(rgb[y*RGB*xsize]), xsize)) {
            /* some sort of failure (already reported by getrow); free storage in this case only */
            free(rgb);
            rgb = NULL;
            goto cleanup;
        }
 	}
 
    *width = xsize;
    *height = ysize;
    *hdrdata = rgb;
    ret = 0;
    
cleanup:
    fclose(fp);
    return ret;
}


/* HDR writer */

#define MINELEN  8            /* minimum scanline length for encoding */
#define MINRUN   4            /* minimum run length */
#define MAXGSHIFT   15        /* maximum shift for gamma table */

/* put out a format value */
static void fputformat(const char *s, FILE *fp)
{
    char  FMTSTR[] = "FORMAT=";
    fputs(FMTSTR, fp);
    fputs(s, fp);
    putc('\n', fp);
}

/* make 4-byte COLR representation from RGB floats */
void setcolr(COLR clr, const float r, const float g, const float b)
{
	int  e;
	double d = r > g ? r : g;
	if (b > d) d = b;
    if (d <= 1e-32) {
        clr[RED] = clr[GRN] = clr[BLU] = 0;
        clr[EXP] = 0;
        return;
    }
	d = frexp(d, &e) * 255.9999 / d;
	clr[RED] = (r > 0.0) ? r * d : 0;
    clr[GRN] = (g > 0.0) ? g * d : 0;
    clr[BLU] = (b > 0.0) ? b * d : 0;
    clr[EXP] = e + COLXS;
    return;
}

/* write out a colr scanline */
static int fwritecolrs(COLR *scanline,
                       const int len,
                       FILE *fp)
{
    int i, j, beg, c2, cnt = 0;
    
    if (len < MINELEN) {        /* too small to encode */
        return ((int) fwrite((char*)scanline, sizeof(COLR), len, fp) - len);
    }
    if (len > 32767) {            /* too big! */
        return (-1);
    }
    putc(2, fp);                /* put magic header */
    putc(2, fp);
    putc(len >> 8, fp);
    putc(len & 255, fp);
    /* put components seperately */
    for (i = 0; i < 4; i++) {
        for (j = 0; j < len; j += cnt) {        /* find next run */
            for (beg = j; beg < len; beg += cnt) {
                for (cnt = 1; cnt < 127 && beg + cnt < len &&
                     scanline[beg + cnt][i] == scanline[beg][i]; cnt++) ;
                if (cnt >= MINRUN)
                    break;      /* long enough */
            }
            if (beg - j > 1 && beg - j < MINRUN) {
                c2 = j + 1;
                while (scanline[c2++][i] == scanline[j][i])
                    if (c2 == beg) {    /* short run */
                        putc(128 + beg - j, fp);
                        putc(scanline[j][i], fp);
                        j = beg;
                        break;
                    }
            }
            while (j < beg) {   /* write out non-run */
                if ((c2 = beg - j) > 128)
                    c2 = 128;
                    putc(c2, fp);
                    while (c2--)
                        putc(scanline[j++][i], fp);
                        }
            if (cnt >= MINRUN) {/* write out run */
                putc(128 + cnt, fp);
                putc(scanline[beg][i], fp);
            } else
                cnt = 0;
                }
    }
    return (ferror(fp) ? -1 : 0);
}


static int radiance_write_header(FILE *fp,
                               const int x,
                               const int y)
{
    const char *COLRFMT = "32-bit_rle_rgbe";
    fprintf(fp, "#?RADIANCE\x0a");
    fprintf(fp, "#HDR from JPEG-HDR\x0a");
    fputformat(COLRFMT, fp);
    fputc('\x0a', fp);
    fprintf(fp, "-Y %d +X %d\x0a", y, x);
    fflush(fp);
    return 1;
}


int write_radiance_hdr(const char* outfile,
                     float* hdr,                 /* interleaved RGB or RGBA float image */
                     const unsigned int stride,  /* 3 for RGB, 4 for RGBA               */
                     const unsigned int width,   /* image dimensions                    */
                     const unsigned int height)
{
    int ret = 1;
    COLR *scanline;
    unsigned int x, y;
    FILE *ofp = fopen(outfile, "wb");
    
    if(NULL == ofp) {
        return ret;
    }

    if ((scanline = (COLR*)calloc(width, sizeof(COLR))) == NULL) {
        goto cleanup;
    }
    if (!radiance_write_header(ofp, width, height)) {
        goto cleanup;
    }
    for (y = 0; y < height; y++) {
        for (x = 0; x < width; x++, hdr+=stride) {
            setcolr(scanline[x], hdr[0], hdr[1], hdr[2]);
        }
        if (fwritecolrs(scanline, width, ofp) < 0) {
            goto cleanup;
        }
    }
    ret = 0;
cleanup:
    free((char*)scanline);
    fclose(ofp);
    return ret;
}


/* 
 The Radiance Software License, Version 1.0
 
 Copyright (c) 1990 - 2002 The Regents of the University of California,
 through Lawrence Berkeley National Laboratory.   All rights reserved.
 
 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions
 are met:
 
 1. Redistributions of source code must retain the above copyright
 notice, this list of conditions and the following disclaimer.
 
 2. Redistributions in binary form must reproduce the above copyright
 notice, this list of conditions and the following disclaimer in
 the documentation and/or other materials provided with the
 distribution.
 
 3. The end-user documentation included with the redistribution,
 if any, must include the following acknowledgment:
 "This product includes Radiance software
 (http://radsite.lbl.gov/)
 developed by the Lawrence Berkeley National Laboratory
 (http://www.lbl.gov/)."
 Alternately, this acknowledgment may appear in the software itself,
 if and wherever such third-party acknowledgments normally appear.
 
 4. The names "Radiance," "Lawrence Berkeley National Laboratory"
 and "The Regents of the University of California" must
 not be used to endorse or promote products derived from this
 software without prior written permission. For written
 permission, please contact radiance@radsite.lbl.gov.
 
 5. Products derived from this software may not be called "Radiance",
 nor may "Radiance" appear in their name, without prior written
 permission of Lawrence Berkeley National Laboratory.
 
 THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 DISCLAIMED.   IN NO EVENT SHALL Lawrence Berkeley National Laboratory OR
 ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 SUCH DAMAGE.
 */
