/******************************************************************************
 * The copyright in this software is being made available under the BSD        *
 * License, included below. This software may be subject to other third party  *
 * and contributor rights, including patent rights, and no such rights are     *
 * granted under this license.                                                 *
 *    <OWNER> = ISO                                                            *
 *    <ORGANIZATION> = ISO                                                     *
 *    <YEAR> = 2015                                                            *
 *    Copyright (c) 2012-2015, ISO                                             *
 *                                                                             *
 * All rights reserved.                                                        *
 * Redistribution and use in source and binary forms, with or without          *
 * modification, are permitted provided that the following conditions are met: *
 *       * Redistributions of source code must retain the above copyright      *
 *         notice,this list of conditions and the following disclaimer.        *
 *       * Redistributions in binary form must reproduce the above             *
 *         copyright notice, this list of conditions and the following         *
 *         disclaimer in the documentation and/or other materials provided     *
 *         with the distribution.                                              *
 *       * Neither the name of the <ORGANIZATION> nor the names of its         *
 *         contributors may be used to endorse or promote products derived     *
 *         from this software without specific prior written permission.       *
 *                                                                             *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" *
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE   *
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE  *
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE   *
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR         *
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF        *
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS    *
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN     *
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)     *
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE  *
 * POSSIBILITY OF SUCH DAMAGE.                                                 *
 ******************************************************************************/

#include <cstdio>
#include <cstdlib>
#include <memory>
#include <cmath>

#include "jpeghdr_operations.h"
#include "jh_helpers.h"
#include "jh_checksum.h"
#include "jh_decoder.h"
#include "jh_encoder.h"
#ifndef WIN32
#include <unistd.h>    /* needed for close() in jh_memmap.h when compiling as C++ */
#endif
#include "jh_helpers.h"
#include "jh_stream.h"
#include "jh_marker_list.h"
#include "file_stream.h"
#include "memmap.h"
#include "hdrio.h"
#include "pfmio.h"
#include "parameters_parser.h"
#include "tonemap.h"

#define VALIDATE_ERROR(err, return_value)  if(null == err) { return (return_value); }
#define CHECK_ERROR(err, return_value) if(JH_OK != err->code) { return (return_value); }

/* Input files are expected to have 3 letter extensions preceded by a '.' separator. */
static eImageFileFormat fileFormatFromExtension(const char * filename)
{
    const unsigned long int iExt = strlen(filename) - 3L;
    if((filename[iExt]=='H') || (filename[iExt]=='h')) {
        return(eHDRFileFormat);
    }
    else if((filename[iExt]=='P') || (filename[iExt]=='p')) {
        if((filename[iExt+1]=='F') || (filename[iExt+1]=='f')) {
            return(ePFMFileFormat);
        }
        else {
            return(eUnknownFormat);
        }
    }
    else if((filename[iExt]=='J') || (filename[iExt]=='j')) {
        return(eJPGFileFormat);
    }
    return(eUnknownFormat);
}


/* tonemapping */

static int jpghdr_tonemap(const jh_image_t* hdri,
                          jh_image_t* sdri,
                          const JPEGOpParams_t& p,
                          jh_error_t *err)
{
    tmo_params_t tmoParams;
    memset(&tmoParams, 0, sizeof(tmo_params_t));
    
    /* get Dolby encoding parameters from file if present */
    if(NULL == p.parametersFile) {
        GetDefaultTonemappingParameters(&tmoParams, err);
    }
    else {
        GetTonemappingParameters(&tmoParams, p.parametersFile, err);
    }
    CHECK_ERROR(err, 1);
    
    jh_monitor_t monitor = { 0 };
    tonemap(&tmoParams, &monitor, hdri, sdri, err);
    CHECK_ERROR(err, 1);
    return 0;
}


/* encoding */

/* write the passed profile data blob into an APP2 marker, tagging the base image's color space */
static jh_marker_node_t* create_icc_profile_marker_node(const unsigned char *icc_profile_data,
                                                     const size_t icc_profile_size)
{
    const unsigned int JPEG_APP2 = JPEG_APP0 + 2;       /* JPEG ICC marker code */
    const unsigned int ICC_HEADER_LEN = 14;             /* size of icc header data in JPEG APP2 markers */
    unsigned char icc_hdr[ICC_HEADER_LEN] = {0x49, 0x43, 0x43, 0x5F, 0x50, 0x52, 0x4F, 0x46, 0x49, 0x4C, 0x45, 0x0, 0x0, 0x0};
    
    const unsigned short icc_length = icc_profile_size + ICC_HEADER_LEN;
    unsigned char *icc_data = (unsigned char *) malloc(icc_length);
    if(NULL == icc_data) {
        std::printf("Failed to allocate memory for ICC APP2 marker; returning NULL instead.\n");
        return NULL;
    }
    memcpy(icc_data, icc_hdr, ICC_HEADER_LEN);
    memcpy(&(icc_data[ICC_HEADER_LEN]), icc_profile_data, icc_profile_size);
    jh_marker_node_t *icc_marker = jh_create_marker_node(JPEG_APP2, icc_length, icc_data);
    free(icc_data);
    return(icc_marker);
}

static int jpghdr_encode(jh_image_t* hdri,
                         jh_image_t* sdri,
                         const JPEGOpParams_t p,
                         const char *outputImage,
                         jh_error_t *err)
{
    int ret = 1;
    jh_encoder_params_t encoder_params = { 0 };
    jh_monitor_t monitor = {0};
    jh_buf_mgr_t mgr = {0};
    jh_file_stream_t strm = {0};
    unsigned char *icc_profile_data = NULL;
    size_t icc_profile_size = 0;
    jh_marker_list_impl_t markers;
    unsigned char *buf = NULL;
    
    VALIDATE_ERROR(err, ret);
    
    /* get Dolby encoding parameters from file if present */
    if(!jh_encoder_set_default(&encoder_params, err)) {
        goto cleanup;
    }
    if(NULL != p.parametersFile) {
        if(GetEncoderParameters(&encoder_params, p.parametersFile, err)) {
            goto cleanup;
        }
    }
    
    /* allow override of file's quality parameters with values from command line */
    if(p.base_quality >= 0) {
        encoder_params.base_quality = p.base_quality;
    }
    if(p.subband_quality >= 0) {
        encoder_params.subband_quality = p.subband_quality;
    }

    /* if an external base image ICC profile is provided, set up an initial markers list in the encoder params struct to hold it */
    if(NULL != p.parametersFile) {
        GetICCProfile(&icc_profile_data, &icc_profile_size, p.parametersFile, err);
        if(err->code != JH_OK) {
            goto cleanup;
        }
    }
    if(icc_profile_size) {  /* if profile successfully read into icc_profile_data */
        jh_init_marker_list(&markers);
        markers.cur = create_icc_profile_marker_node(icc_profile_data, icc_profile_size);
        encoder_params.markers = &(markers.pub);
    }
    
    /* set up file stream for output file outPath */
    buf = (unsigned char*) malloc(JH_FBUF_SIZE);
    if (buf == NULL) {
        jh_set_error(err, JH_ERROR, "Encoder failed to allocate buffer for the stream on file %s\n", outputImage);
        goto cleanup;
    }
    mgr.on_empty = jh_on_fstrm_empt;
    mgr.on_init = jh_on_fstrm_init;
    mgr.on_term = jh_on_fstrm_term;
    strm.pub.write = jh_fstream_write;
    strm.pub.seek = jh_fstream_seek;
    strm.pub.read = jh_fstream_read;   /* needed for checksum computation */
    strm.buf = buf;
    strm.buf_size = JH_FBUF_SIZE;
    strm.pub.pri = &mgr;
    strm.stream = fopen(outputImage, "w+b");
    if(strm.stream == null) {
        jh_set_error(err, JH_ERROR, "Encoder failed to create stream for jpeg-hdr output file %s\n", outputImage);
        goto cleanup;
    }
    ret = jh_encode (&encoder_params, &monitor, hdri, sdri, (jh_stream_t*)&strm, err);
    if (err->code != JH_OK) {
        goto cleanup;
    }
    else if (JH_OK != jh_add_checksum((jh_stream_t*)&strm, err)) {
        goto cleanup;
    }
    if (strm.stream != null) {
        fclose(strm.stream);
    }
    ret = 0;
cleanup:
    /* clean up any created profile */
    if(icc_profile_size) {
        markers.pub.empty(&markers.pub);
        free(icc_profile_data);
    }
    free(buf);
    return ret;
}

static jh_image_t* read_8bit_ppm(const char *ppmfile, const float *trc256, jh_error_t *err)
{
    jh_image_t* sdri = NULL;
    int i = 0, error = 0;
    char type;
    VALIDATE_ERROR(err, NULL);
    FILE *infile = fopen(ppmfile, "rb");
    if (NULL == infile) {
        jh_set_error(err, JH_ERROR, "Failed to open tonemapped image PPM file %s.", ppmfile);
        return NULL;
    }
    
    unsigned long  width = 0, height = 0;
    int channels = 0;
    char ppmline[256];
    int maxval;

    /* read the file ID */
    fgets(&(ppmline[0]), 256, infile);
    type = ppmline[1];
    if (ppmline[0] != 'P' || ((type != '6') && (type != '5'))) {
        jh_set_error(err, JH_ERROR, "Image %s not a supported PPM file type.", ppmfile);
        error = 1;
        goto cleanup;
    }
    channels = ('6' == type) ? 3 : 1;

    /* skip over comments, then get width and height */
    do {
        fgets(ppmline, 256, infile);
    } while (ppmline[0] == '#');
    sscanf(ppmline, "%lu %lu", &width, &height);
    
    if(0 == height) { /* height on separate line? */
        do {
            fgets(ppmline, 256, infile);
        } while (ppmline[0] == '#');
        sscanf(ppmline, "%lu", &height);
    }
    
    /* skip over comments, then get max value */
    do {
        fgets(ppmline, 256, infile);
    } while (ppmline[0] == '#');
    sscanf(ppmline, "%d", &maxval);
    if (maxval != 255) {
        jh_set_error(err, JH_ERROR, "Tonemapped PPM image maxval = %d; expected 255.", maxval);
        error = 1;
        goto cleanup;
    }
    
    sdri = jh_create_image((int) width, (int) height, (1 == channels) ? JH_GRAY8 : JH_RGB8);
    if(NULL == sdri) {
        jh_set_error(err, JH_ERROR, "Failed to allocate memory for tonemapped image struct.");
        error = 1;
        goto cleanup;
    }
    
    /* overwrite decoding LUT with passed table; jh_create_image() allocates a 256-entry table for 8-bit pixel types */
    if (JH_CS_TRC_SIZE > sdri->lut_size) {
        jh_set_error(err, JH_ERROR, "Storage too small for tonemapped image decoding LUT");
        error = 1;
        goto cleanup;
    }
    sdri->lut_size = JH_CS_TRC_SIZE;
    memcpy(sdri->decoding_lut, trc256, sdri->lut_size * sizeof(float));
    
    /* do row-wise read, honoring any padding in sdri's rows */
    for(i=0; i<sdri->rows; i++) {
        unsigned char *prow = (unsigned char *) jh_get_row(sdri, i);
        if(sdri->stride != fread(prow, sizeof(unsigned char), sdri->stride, infile)) {
            jh_set_error(err, JH_ERROR, "Failed to read tonemapped image data.");
            error = 1;
            goto cleanup;
        }
    }
cleanup:
    fclose(infile);
    if(error) {
        jh_delete_image(sdri);
        return NULL;
    }
    return sdri;
}

jh_image_t * ReadTonemappedImage(const char *tonemappedImage, const JPEGOpParams_t& p, jh_error_t *err)
{
    jh_encoder_params_t encoder_params;
    VALIDATE_ERROR(err, null);
    /* get Dolby encoding parameters from file if present */
    jh_encoder_set_default(&encoder_params, err);
    CHECK_ERROR(err, null);
    if(NULL != p.parametersFile) {
        GetEncoderParameters(&encoder_params, p.parametersFile, err);
        CHECK_ERROR(err, null);
    }
    return(read_8bit_ppm(tonemappedImage, encoder_params.base_cs_trc, err));
}

int Encode(const char *inputImage,
           const char *outputImage,
           const char *tonemappedImage,  /* may be NULL; invokes built-in exposure-gamma TMO */
           const JPEGOpParams_t& p,
           jh_error_t *err)
{
    unsigned int rows=0U, cols=0U, channels=0U;
    float *hdr = NULL;
    jh_image_t *hdri=NULL, *sdri=NULL;
    const unsigned int RGB = 3;
    int i, ret = 1, errcode = 0;
    VALIDATE_ERROR(err, ret);
    switch(fileFormatFromExtension(inputImage)) {
        case ePFMFileFormat:
            errcode = read_pfm(inputImage, &hdr, &cols, &rows, &channels);
            break;
        case eHDRFileFormat:
            errcode = read_radiance_hdr(inputImage, &hdr, &cols, &rows);
            channels = RGB;
            break;
        default:
            jh_set_error(err, JH_ERROR, "Missing file or unknown file format for input image %s.", inputImage);
            return ret;
    }
    if (errcode) {
        jh_set_error(err, JH_ERROR, "Failed to read input image %s (reader error code %d).\n", inputImage, errcode);
        goto cleanup;
    }
    hdri = jh_create_image(cols, rows, (RGB == channels) ? JH_RGBF : JH_GRAYF);
    if (NULL == hdri) {
        jh_set_error(err, JH_ERROR, "Failed to allocate memory for HDR image.");
        goto cleanup;
    }
    /* copy raster data into HDR image struct, row-wise to account for possible padding */
    for(i=0; i<rows; i++) {
        memcpy(jh_get_row(hdri, i), &(hdr[i*cols*channels]), hdri->stride);
    }
    if(NULL == tonemappedImage) {
        sdri = jh_create_image(cols, rows, (RGB == channels) ? JH_RGB8 : JH_GRAY8);
        if(NULL == sdri) {
            jh_set_error(err, JH_ERROR, "Failed to allocate memory for tonemapped image.");
            goto cleanup;
        }
        if(jpghdr_tonemap(hdri, sdri, p, err)) {
            goto cleanup;
        }
    }
    else {
        /* read the tonemapped image */
        sdri = ReadTonemappedImage(tonemappedImage, p, err);
        if(NULL == sdri) {
            goto cleanup;
        }
        
        /* check that dimensions are correct */
        if((hdri->rows != sdri->rows) || (hdri->columns != sdri->columns)) {
            jh_set_error(err, JH_ERROR, "Mismatched HDR and tonemapped image dimensions." );
            goto cleanup;
        }
        
        /* check that channel counts agree */
        if (jh_channel_count(hdri->pixel_type) != jh_channel_count(sdri->pixel_type) ) {
            jh_set_error(err, JH_ERROR, "Mismatched HDR and tonemapped image channel counts." );
            goto cleanup;
        }
    }
    ret = jpghdr_encode(hdri, sdri, p, outputImage, err);
    
cleanup:
    jh_delete_image(hdri);
    jh_delete_image(sdri);
    free(hdr);
    return ret;
}

/* decoding */

static jh_image_t * jpghdr_decode(const char* imagePath, const JPEGOpParams_t& p, jh_error_t *err)
{
    jh_decoder_params_t dp = {0};
    volatile int cancel = 0;
    volatile float progress = 0;
    jh_monitor_t monitor = {cancel, progress};

    VALIDATE_ERROR(err, null);
    /* get Dolby decoding parameters from file if present */
    jh_decoder_set_default(&dp, err);
    CHECK_ERROR(err, null);
    if(NULL != p.parametersFile) {
        GetDecoderParameters(&dp, p.parametersFile, err);
        CHECK_ERROR(err, null);
    }
    
    /* cannot have a NULL marker list in the decoder params struct */
    jh_marker_list_impl_t markers = { 0 };  /* N.B.: this initialization is required */
    jh_init_marker_list(&markers);
    dp.marker_list = &(markers.pub);
    
    size_t size = 0;
    void* buf = memmap(imagePath, &size);
    if (buf == null) {
        jh_set_error(err, JH_ERROR, "Decoder failed to map file %s", imagePath);
        return null;
    }
    jh_mem_stream_t strm = {0};
    jh_buf_mgr_t mgr = {0};
    jh_init_stream(&strm, &mgr);
    jh_attach_membuf(&strm, (unsigned char*)buf, size, size); /* take ownership of the buffer */
    if(JH_OK != jh_is_checksum_valid((jh_stream_t*)&strm, err)) {
        if((JH_INVALID_CHECKSUM != err->code) && (JH_HEADER_NO_CHECKSUM != err->code)) {
            /* a more serious error occurred, possibly a failure to rewind the stream. */
            return null;
        }
        else {
            /* otherwise, inform the user, then reset the error struct so we may do the decoding */
            fprintf(stderr, "WARNING: Invalid or missing checksum in image %s; attempting to continue decoding.\n", imagePath);
            jh_set_error(err, JH_OK, "No error.");
        }
    }
    jh_image_t* hdri = jh_decode((jh_stream_t*)&strm, &monitor, &dp, err);
    memunmap(buf, size);
    CHECK_ERROR(err, null);
    jh_empty_stream((jh_stream_t*) &strm);
    markers.pub.empty(&markers.pub);
    return hdri;
}

int Decode(const char *inputImage,
           const char *outputImage,
           const JPEGOpParams_t& p,
           jh_error_t *err)
{
    const unsigned int RGB = 3U;
    unsigned int channels = RGB;
    int ret = 1;
    int errcode = 0;
    VALIDATE_ERROR(err, ret);
    jh_image_t* hdri = jpghdr_decode(inputImage, p, err);
    if (null == hdri) {
        return ret;
    }
    if(err->code != JH_OK) {
        goto cleanup;
    }
    if ((JH_RGBF != hdri->pixel_type) && (JH_GRAYF != hdri->pixel_type)) {
        jh_set_error(err, JH_ERROR, "Decoded JPEG-HDR image does not have expected float pixel type.");
        goto cleanup;
    }
    if(JH_GRAYF == hdri->pixel_type) {
        channels = 1U;
    }
    switch(fileFormatFromExtension(outputImage)) {
        case ePFMFileFormat:
            errcode = write_pfm(outputImage, (float *) hdri->data, channels, (unsigned int) hdri->columns, (unsigned int) hdri->rows);
            break;
        case eHDRFileFormat:
            if(JH_RGBF == hdri->pixel_type) {
                errcode = write_radiance_hdr(outputImage, (float *) hdri->data, RGB, (unsigned int) hdri->columns, (unsigned int) hdri->rows);
            }
            else {
                /* radiance format doesn't support gray, so convert to RGB via duplication and write */
                float *rgbData = (float *) malloc(RGB * hdri->columns * hdri->rows * sizeof(float));
                if(NULL == rgbData) {
                    jh_set_error(err, JH_ERROR, "Failed to allocate memory for converting decoded gray image to RGB for writing HDR image %s.", outputImage);
                    goto cleanup;
                }
                for(unsigned int k=0, i=0; i<hdri->rows; i++) {
                    float *grayData = (float *) jh_get_row(hdri, i);
                    for(unsigned int j=0; j<hdri->columns; j++) {
                        rgbData[k++] = grayData[j];
                        rgbData[k++] = grayData[j];
                        rgbData[k++] = grayData[j];
                    }
                }
                errcode = write_radiance_hdr(outputImage, rgbData, RGB, (unsigned int) hdri->columns, (unsigned int) hdri->rows);
                free(rgbData);
            }
            break;
        default:
            break;
    }
    if (errcode) {
        jh_set_error(err, JH_ERROR, "Failed to write output image %s.", outputImage, errcode);
        goto cleanup;
    }
    ret = 0;
cleanup:
    jh_delete_image(hdri);
    return ret;
}

#undef VALIDATE_ERROR
#undef CHECK_ERROR
